
; the following arguments (divident, divisor, remainder and division
; counter are located in the bank 15
DVDT_0	equ	0xf00
DVDT_1	equ	0xf01
DVDT_2	equ	0xf02
DVDT_3	equ	0xf03				; dividend

DVSR_0	equ	0xf04
DVSR_1	equ	0xf05
DVSR_2	equ	0xf06
DVSR_3	equ	0xf07				; divisor

RMDR_0	equ	0xf08
RMDR_1	equ	0xf09
RMDR_2	equ	0xf0a
RMDR_3	equ	0xf0b				; remainder

DIVCTR	equ	0xf0c

; ------------------------- Multiply two 16 bit variables (words) -------------------------------
; in: DVDT=argument1, DVSR=argument2
; out: RMDR=result
; arguments are located in the page 15
MUL16:
	movlb	0x0f
	clrf	RMDR_0,BANKED
	clrf	RMDR_1,BANKED
	clrf	RMDR_2,BANKED
	clrf	RMDR_3,BANKED		; clear result

	movlw	0x10				; 16 bits
	movwf	DIVCTR,BANKED
MUL16_Cycle:
	rrcf	DVDT_3,1,BANKED
	rrcf	DVDT_2,1,BANKED
	rrcf	DVDT_1,1,BANKED
	rrcf	DVDT_0,1,BANKED		; rotate divident right to C
	bnc		MUL16_NoAdd

	; adding DVSR to RMDR
	call	Add16_RMDR_eq_DVSR_plus_RMDR
MUL16_NoAdd:
	decfsz	DIVCTR,1,BANKED
	bra		MUL16_Cycle
	return

; ------------------------- Divide two 32 bit variables (dwords) -------------------------------
; in: DVDT=dividend, DVSR=divisor
; out: DVDT=result, RMDR=remainder
; arguments are located in the page 15
DIV32:
	movlb	0x0f
	clrf	RMDR_0,BANKED
	clrf	RMDR_1,BANKED
	clrf	RMDR_2,BANKED
	clrf	RMDR_3,BANKED		; clear remainder

	movlw	0x20				; 32 bits
	movwf	DIVCTR,BANKED
	rlcf	DVDT_0,1,BANKED
	rlcf	DVDT_1,1,BANKED
	rlcf	DVDT_2,1,BANKED
	rlcf	DVDT_3,1,BANKED		; rotate divident left to C
DIV32_Loop:
	rlcf	RMDR_0,1,BANKED		; rotate C into LSB of remainder
	rlcf	RMDR_1,1,BANKED
	rlcf	RMDR_2,1,BANKED
	rlcf	RMDR_3,1,BANKED		; rotate remainder left to C
	bc		DIV32_substract

	call	DIV32_subs_rmdr_dvsr_nc
	bnc		DIV32_NoAdd

	; now add it back - substraction result was negative
	movf	DVSR_0,W,BANKED
	addwf	RMDR_0,1,BANKED
	movf	DVSR_1,W,BANKED
	addwfc	RMDR_1,1,BANKED	
	movf	DVSR_2,W,BANKED
	addwfc	RMDR_2,1,BANKED
	movf	DVSR_3,W,BANKED
	addwfc	RMDR_3,1,BANKED

DIV32_NoAdd:
	btg		STATUS,C,ACCESS		; invert C flag

DIV32_shift_divident:
	rlcf	DVDT_0,1,BANKED
	rlcf	DVDT_1,1,BANKED
	rlcf	DVDT_2,1,BANKED
	rlcf	DVDT_3,1,BANKED		; rotate divident left to C
	decfsz	DIVCTR,1,BANKED
	bra		DIV32_Loop
	return

DIV32_substract:
	call	DIV32_subs_rmdr_dvsr_nc
	bra		DIV32_shift_divident

DIV32_subs_rmdr_dvsr_nc:
	; substract divisor from remainder, returns C flag
	movf	DVSR_0,W,BANKED
	subwf	RMDR_0,1,BANKED
	movf	DVSR_1,W,BANKED
	subwfb	RMDR_1,1,BANKED
	movf	DVSR_2,W,BANKED
	subwfb	RMDR_2,1,BANKED
	movf	DVSR_3,W,BANKED
	subwfb	RMDR_3,1,BANKED
	btg		STATUS,C,ACCESS			; toggle C flag bit as it is inversed after substraction op
	return

; ---------------------------- Divide two 16 bit variables (words) -----------------------------
; in: DVDT=dividend, DVSR=divisor
; out: DVDT=result, RMDR=remainder
; arguments are located in the page 15
DIV16:
	movlb	0x0f				; page 15
	clrf	RMDR_0,BANKED
	clrf	RMDR_1,BANKED		; clear remainder

	movlw	0x10				; 16 bits
	movwf	DIVCTR,BANKED
	rlcf	DVDT_0,1,BANKED
	rlcf	DVDT_1,1,BANKED		; rotate divident left to C
DIV16_Loop:
	rlcf	RMDR_0,1,BANKED		; rotate C into LSB of remainder
	rlcf	RMDR_1,1,BANKED		; rotate remainder left to C
	bc		DIV16_substract
	call	DIV16_subs_rmdr_dvsr_nc
	bnc		DIV16_NoAdd

	; now add with no carry
	call	Add16_RMDR_eq_DVSR_plus_RMDR

DIV16_NoAdd:
	btg		STATUS,C,ACCESS		; invert C flag

DIV16_shift_divident:
	rlcf	DVDT_0,1,BANKED
	rlcf	DVDT_1,1,BANKED		; rotate divident left to C
	decfsz	DIVCTR,1,BANKED
	bra		DIV16_Loop
	return

DIV16_substract:
	call	DIV16_subs_rmdr_dvsr_nc
	bra		DIV16_shift_divident

DIV16_subs_rmdr_dvsr_nc:
	; substract divisor from remainder, returns C flag
	movf	DVSR_0,W,BANKED
	subwf	RMDR_0,1,BANKED
	movf	DVSR_1,W,BANKED
	subwfb	RMDR_1,1,BANKED	
	btg		STATUS,C,ACCESS			; toggle C flag bit as it is inversed after substraction op
	return

Add16_RMDR_eq_DVSR_plus_RMDR:
	movf	DVSR_0,W,BANKED
	addwf	RMDR_0,1,BANKED
	movf	DVSR_1,W,BANKED
	addwfc	RMDR_1,1,BANKED	
	return

HDDRST:
	movlw	0x90
	call	WriteHDDCommand
	return

; --- Returns selected HDD status to host
HDDSTAT:
; in: bit 0 drive number (0/1)
; out: status register of specified drive
	movlw	0xAB					; command accepted: now stage 2 to get first parameter from host
	call	WaitForHostRead			; return value
	btfss	STATUS,C,ACCESS			; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	call	WaitForHostWrite		; hard drive ID (0/1 = master/slave)
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	swapf	WREG,W,ACCESS
	andlw	0x10					; drive ID paramter in drv/head register format
	call	WriteHDDDrvHead

	; getting status register of specified drive
	call	ReadHDDStatus
	movf	LATA,W,ACCESS
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	call	ReadHDDError
	movf	LATA,W,ACCESS
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

; --- Get HDD ID
; in: parameter bit 0: drive (0/1), bit 7: 0 (512 bytes ID information), 1 (only CHS, model and capabiliries)
HDDID:
	movlw	0xAB					; command accepted: now stage 2 to get first parameter from host
	call	WaitForHostRead			; return value
	btfss	STATUS,C,ACCESS			; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	call	WaitForHostWrite		; hard drive ID (0/1 = master/slave)
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	movwf	FMTPR2,ACCESS			; bit 7 = return only main disk parameters
	andlw	0x01
	movwf	FMTPRM,ACCESS

	; set current drive
	swapf	FMTPRM,W,ACCESS
	andlw	0x10					; 0x00 for master drive and 0x10 for slave drive
	call	WriteHDDDrvHead
	call	HDDWaitReady
	btfsc	STATUS,C,ACCESS
	bra		HDDIORetStat

	movlw	0x00
	call	WriteHDDSectorCount
	movlw	0x00
	call	WriteHDDCylHigh
	movlw	0x00
	call	WriteHDDCylLow
	movlw	0x00
	call	WriteHDDSectorNumber
	movlw	0xEC					; Identify command
	call	WriteHDDCommand

	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS
HDDDataWaitLoop:
	call	ReadHDDStatus
	btfsc	LATA,7,ACCESS
	bra		HDDDataWaitRetry
	btfsc	LATA,3,ACCESS
	bra		HDDIDDataReady
HDDDataWaitRetry:
	decfsz	TIMOUT1,1,ACCESS
	bra		HDDDataWaitLoop
	decfsz	TIMOUT2,1,ACCESS
	bra		HDDDataWaitLoop
	bra		HDDIORetStat

HDDIDDataReady:
	movlw	0x01
	movwf	FSR0H,ACCESS	
	clrf	FSR0L,ACCESS			; buffer start at 0x100

	clrf	FMTPR1,ACCESS			; counter clear, 256 counts
ReadHDDData:
	call	ReadHDDDataReg
	movf	LATA,W,ACCESS
	movwf	POSTINC0,ACCESS
	movf	LATB,W,ACCESS
	movwf	POSTINC0,ACCESS
	decfsz	FMTPR1,1,ACCESS
	bra		ReadHDDData

	call	HDDWaitReady
	bc		HDDIORetStat

	btfss	FMTPR2,7,ACCESS			; bit 7 of input parameter
	bra		HDDIDOK					; if it's 0, do not perform formatted output

	; returns the following information:
	; number of cylinders (2 bytes)
	; number of heads (1 byte)
	; number of sectors per track (1 byte)
	; model number, string, terminated with 0x00 (max 40 bytes)
	; capabilities flag (1 byte): bit 0 - DMA support, bit 1 - LBA support

	; prepare model string
	movlw	0x1b*2
	movwf	FSR0L,ACCESS
	movlw	0x01
	movwf	FSR0H,ACCESS
	movlw	0x14
	movwf	FMTPR1,ACCESS
IDSwapBytes:
	movf	POSTINC0,W,ACCESS
	movff	POSTDEC0,POSTINC0
	movwf	POSTINC0,ACCESS
	decfsz	FMTPR1,1,ACCESS
	bra		IDSwapBytes

	; check string from the end
	movlw	0x1b*2					; LOW start of the ID string (0x36)
	movwf	FMTPR1,ACCESS
	movlw	0x01
	movwf	FSR0H,ACCESS
	movlw	0x2e*2+1
	movwf	FSR0L,ACCESS			; end of ID string 0x15D
HIDPrevChar:
	movf	INDF0,W,ACCESS
	bz		HIDTestEnd
	sublw	0x20
	bnz		HIDStrEnd
HIDTestEnd:
	decf	FSR0L,1,ACCESS
	movf	FSR0L,W,ACCESS
	subwf	FMTPR1,W,ACCESS			; compare to the beginning of the string
	bnz		HIDPrevChar
	bra		IDEndOfStrSend			; reached beginning of the ID string buffer - it is empty

HIDStrEnd:
	movf	FSR0L,W,ACCESS
	movwf	FMTPR1,ACCESS			; save position of the last valid char in the ID string
	movlw	0x1b*2
	movwf	FSR0L,ACCESS			; now points to the beginning of the ID string

	; proceed outputting to the host
	movlw	0xAE
	call	WaitForHostRead

	movlb	1						; absolute addresses ???
	; number of cylinders
	movf	0x02,W,BANKED
	call	WaitForHostRead
	movf	0x03,W,BANKED
	call	WaitForHostRead
	; number of heads
	movf	0x06,W,BANKED
	call	WaitForHostRead
	; number of sectors per track
	movf	0x0C,W,BANKED
	call	WaitForHostRead

	; model string
IDStrOut:
	movf	POSTINC0,W,ACCESS
	bz		IDEndOfStrSend			; inexpected 0x00 - terminate output
	call	WaitForHostRead			; WREG is preserved!
	movf	FSR0L,W,ACCESS
	subwf	FMTPR1,W,ACCESS
	bc		IDStrOut				; keep writing bytes until reached end of valid ID string
									; Carry is negated in substraction command
IDEndOfStrSend:
	movlw	0x00
	call	WaitForHostRead			; write end of string

	; capabilities
	movf	0x63,W,BANKED
	call	WaitForHostRead
	movlb	0

HDDIDOK:
	movlw	0xCC

HDDIORetStat:
	movwf	SECPRM,ACCESS			; preserve return byte
	movlw	0xAF
	call	WaitForHostRead			; returning 0xAF - result phase
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	SECPRM,W,ACCESS			; restore return byte
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

; *** wait for HDD to become ready (BUSY=0, RDY=1)
HDDWaitReady:
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS		; long timeout = 0x10000 loops

HDD_WaitReady:
	nop
	nop								; 125 ns delay
	call	ReadHDDStatus
	movf	LATA,W,ACCESS
	andlw	0xC0
	sublw	0x40					; BUSY (bit 7)=0, RDY (bit 6)=1
	bz		HDDBecameReady

	decfsz	TIMOUT1,1,ACCESS
	bra		HDD_WaitReady
	decfsz	TIMOUT2,1,ACCESS
	bra		HDD_WaitReady

	bsf		STATUS,C,ACCESS			; raise C flag = timeout error
	return

HDDBecameReady:
	bcf		STATUS,C,ACCESS			; reset C flag
	return							; returns C flag=0

; *** wait for HDD to become for data transfer (BUSY=0, DRQ=1)
HDDWaitData:
	clrf	TMP,ACCESS			; flag that DRQ happened
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS		; long timeout = 0x10000 loops

HDD_WaitData:
	call	ReadHDDStatus
	btfsc	LATA,0,ACCESS
	bra		DataWaitRetErr		; if error bit is set

	btfsc	LATA,3,ACCESS		; DRQ is set?
	setf	TMP,ACCESS			; set flag that data is ready
	btfss	LATA,7,ACCESS		; BUSY is set?
	bra		HDDCheckDRQ			; not busy, go check if DRQ happened

	decfsz	TIMOUT1,1,ACCESS
	bra		HDD_WaitData
	decfsz	TIMOUT2,1,ACCESS
	bra		HDD_WaitData
DataWaitRetErr:
	bsf		STATUS,C,ACCESS			; raise C flag = timeout error
	return

HDDCheckDRQ:
	bsf		STATUS,C,ACCESS			; set C flag
	btfsc	TMP,0,ACCESS			; was DRQ set some time during loop?
	bcf		STATUS,C,ACCESS			; yes, it was - reset C flag
	return							; returns C flag=0

; --- Writes to HDD data register from LATA/LATB
WriteHDDDataReg:
; in: LATA=low byte, LATB=high byte
	movlw	0xb8				; setting up /HDDCS=0, Address=0
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 125 ns
	nop
	nop							; 187.5ns /WR pulse width (16-bit write)
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS		; invalidating address and inactivating /HDDCS
	return

; --- Reads HDD data register to LATA/LATB
ReadHDDDataReg:
; out: LATA=low byte, LATB=high byte
	movlw	0xb8				; setting up /HDDCS=0, Address=0
	movwf	PORTC,ACCESS
	setf	TRISA,ACCESS		; port A = input
	setf	TRISB,ACCESS		; port B = input
	bcf		LATC,3,ACCESS		; setting up /RD in 62.5 ns
	nop
	nop
	movf	PORTA,1,ACCESS		; write input of port A into LATA
	movf	PORTB,1,ACCESS		; write input of port B into LATB
	bsf		LATC,3,ACCESS		; inactivating /RD
	clrf	TRISA,ACCESS		; port A = output
	clrf	TRISB,ACCESS		; port B = output
	setf	PORTC,ACCESS
	return

; --- Reads HDD error register to LATA
ReadHDDError:
; out: LATA=status register
	movlw	0xb9				; setting up /HDDCS=0, Address=1
	movwf	PORTC,ACCESS
	setf	TRISA,ACCESS		; port A = input
	bcf		LATC,3,ACCESS		; setting up /RD in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /RD pulse duration till data read
	movf	PORTA,1,ACCESS		; write input of port A into LATA
	bsf		LATC,3,ACCESS		; inactivating /RD
	clrf	TRISA,ACCESS		; port A = output
	setf	PORTC,ACCESS
	return

; --- Reads HDD status register to LATA
ReadHDDStatus:
; out: LATA=status register
	movlw	0xbf				; setting up /HDDCS=0, Address=7
	movwf	PORTC,ACCESS
	setf	TRISA,ACCESS		; port A = input
	bcf		LATC,3,ACCESS		; setting up /RD in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /RD pulse duration till data read
	movf	PORTA,1,ACCESS		; write input of port A into LATA
	bsf		LATC,3,ACCESS		; inactivating /RD
	clrf	TRISA,ACCESS		; port A = output
	setf	PORTC,ACCESS
	return

; --- Writes WREG to the HDD command register
WriteHDDCommand:
	movwf	PORTA,ACCESS
	movlw	0xbf				; setting up /HDDCS=0, Address=7
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return

; --- Writes WREG to the HDD drive/head register
WriteHDDDrvHead:
	andlw	0x1f			; mask bits 7-5 out
	iorlw	0xa0			; set bits 7 and 5 (bit 6 is reset)
	movwf	PORTA,ACCESS
	movlw	0xbe				; setting up /HDDCS=0, Address=6
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	nop
	nop
	nop
	nop
	nop
	nop
	nop								; 7 NOP commands = 437.5 ns (400ns unit select delay)
	return

; --- Writes WREG to the HDD sector number register
WriteHDDSectorNumber:
	movwf	PORTA,ACCESS
	movlw	0xbb				; setting up /HDDCS=0, Address=3
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return

; --- Reads HDD sector number register to LATA
ReadHDDSectorCount:
; out: LATA=status register
	setf	PORTA,ACCESS
	movlw	0xba				; setting up /HDDCS=0, Address=2
	movwf	PORTC,ACCESS
	setf	TRISA,ACCESS		; port A = input
	bcf		LATC,3,ACCESS		; setting up /RD in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /RD pulse duration till data read
	movf	PORTA,1,ACCESS		; write input of port A into LATA
	bsf		LATC,3,ACCESS		; inactivating /RD
	clrf	TRISA,ACCESS		; port A = output
	setf	PORTC,ACCESS
	return

; --- Writes WREG to the HDD cylinder LOW register
WriteHDDCylLow:
	movwf	PORTA,ACCESS
	movlw	0xbc				; setting up /HDDCS=0, Address=4
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return

; --- Writes WREG to the HDD cylinder HIGH register
WriteHDDCylHigh:
	movwf	PORTA,ACCESS
	movlw	0xbd				; setting up /HDDCS=0, Address=5
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return

; --- Writes WREG to the HDD sector count register
WriteHDDSectorCount:
	movwf	PORTA,ACCESS
	movlw	0xba				; setting up /HDDCS=0, Address=2
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return

; --- Writes WREG to the HDD features register
WriteHDDFeatures:
	movwf	PORTA,ACCESS
	movlw	0xb9				; setting up /HDDCS=0, Address=1
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return